package com.vaadin.incubator.dashlayout.client.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import com.google.gwt.dom.client.Document;
import com.google.gwt.user.client.ui.ComplexPanel;
import com.google.gwt.user.client.ui.Widget;
import com.google.gwt.user.client.ui.WidgetCollection;
import com.vaadin.incubator.dashlayout.client.util.css.CSSRule;
import com.vaadin.incubator.dashlayout.client.util.css.CSSUtil;
import com.vaadin.terminal.gwt.client.ApplicationConnection;
import com.vaadin.terminal.gwt.client.BrowserInfo;
import com.vaadin.terminal.gwt.client.Container;
import com.vaadin.terminal.gwt.client.Paintable;
import com.vaadin.terminal.gwt.client.RenderSpace;
import com.vaadin.terminal.gwt.client.UIDL;
import com.vaadin.terminal.gwt.client.ValueMap;
import com.vaadin.terminal.gwt.client.ui.VMarginInfo;

public class VDashLayout extends ComplexPanel implements Container {

    public static final String CLASSNAME = "v-dashlayout";

    protected boolean horizontal = false;

    /**
     * A logical mapping from widget to its virtual layout cell.
     */
    protected Map<Widget, ChildCell> widgetToCell = new HashMap<Widget, ChildCell>();

    /**
     * All array values are in the following order: top, right, bottom, left.
     */

    /* Current margin values */
    protected int[] margin = { 0, 0, 0, 0 };

    /* Current padding values */
    protected int[] padding = { 0, 0, 0, 0 };

    /* Current border sizes */
    protected int[] border = { 0, 0, 0, 0 };

    /*
     * Current inner size (excluding margins, border and padding) in pixels
     */
    protected int width = -1;
    protected int height = -1;

    protected boolean isRendering;
    protected boolean sizeHasChangedDuringRendering = false;

    // Is the layout size undefined, i.e. defined by the contained components
    protected boolean undefWidth = false;
    protected boolean undefHeight = false;

    protected ApplicationConnection client;

    protected String lastStyleName;

    protected VMarginInfo marginInfo;

    protected float compoundRatio = -1;

    protected int consumedSpace = 0;

    protected int spacing = 0;
    protected boolean useSpacing = true;

    public VDashLayout() {
        super();
        setElement(Document.get().createDivElement());
        setStylePrimaryName(CLASSNAME);
        getElement().getStyle().setProperty("float", "left");
        getElement().getStyle().setProperty("cssFloat", "left");
        getElement().getStyle().setProperty("styleFloat", "left");
        getElement().getStyle().setProperty("display", "inline");
        if (BrowserInfo.get().isIE()) {
            getElement().getStyle().setProperty("zoom", "1");
        }
    }

    public boolean isHorizontal() {
        return horizontal;
    }

    @Override
    public WidgetCollection getChildren() {
        return super.getChildren();
    }

    public int getSpacing() {
        return spacing;
    }

    public float getCompoundRatio() {
        return compoundRatio;
    }

    public Map<Widget, ChildCell> getCells() {
        return widgetToCell;
    }

    public int getConsumedSpace() {
        return consumedSpace;
    }

    public void updateActualSize() {
        width = CSSUtil
                .parsePixel(CSSUtil.getStyleValue(getElement(), "width"));
        height = CSSUtil.parsePixel(CSSUtil.getStyleValue(getElement(),
                "height"));
    }

    private void updateMargins() {
        margin = CSSUtil.collectMargin(getElement());
        if (marginInfo.hasTop()) {
            getElement().getStyle().setProperty("marginTop", "");
        } else {
            getElement().getStyle().setProperty("marginTop", "0");
            margin[0] = 0;
        }
        if (marginInfo.hasRight()) {
            getElement().getStyle().setProperty("marginRight", "");
        } else {
            getElement().getStyle().setProperty("marginRight", "0");
            margin[1] = 0;
        }
        if (marginInfo.hasBottom()) {
            getElement().getStyle().setProperty("marginBottom", "");
        } else {
            getElement().getStyle().setProperty("marginBottom", "0");
            margin[2] = 0;
        }
        if (marginInfo.hasLeft()) {
            getElement().getStyle().setProperty("marginLeft", "");
        } else {
            getElement().getStyle().setProperty("marginLeft", "0");
            margin[3] = 0;
        }
    }

    private void updateDynamicSizeInfo(UIDL uidl) {
        String w = uidl.hasAttribute("width") ? uidl
                .getStringAttribute("width") : "";
        undefWidth = w.equals("");
        String h = uidl.hasAttribute("height") ? uidl
                .getStringAttribute("height") : "";
        undefHeight = h.equals("");
    }

    private void measureCSSAndSpacing() {
        margin = CSSUtil.collectMargin(getElement());
        padding = CSSUtil.collectPadding(getElement());
        border = CSSUtil.collectBorder(getElement());

        if (useSpacing) {
            final String[] styles = getElement().getClassName().split(" ");
            for (int i = styles.length - 1; i >= 0; i--) {
                String style = styles[i];
                final CSSRule r = new CSSRule("#" + style + "-spacing", false);
                final String width = r.getPropertyValue("width");
                if (width != null) {
                    spacing = CSSUtil.parsePixel(width);
                    break;
                }
            }
        } else {
            spacing = 0;
        }
    }

    @Override
    public void setStyleName(String styleName) {
        super.setStyleName(styleName);

        if (isAttached() && !styleName.equals(lastStyleName)) {
            measureCSSAndSpacing();
            // updateChildCells(null, false);
            lastStyleName = styleName;
        }

    }

    public RenderSpace getAllocatedSpace(Widget child) {
        ChildCell cell = widgetToCell.get(child);
        if (child instanceof VDashLayout) {
            // Layout handles it's margins internally
            return cell.getRenderSpace();
        }
        return cell.getSpaceSansMargins();
    }

    public boolean hasChildComponent(Widget component) {
        return getChildren().contains(component);
    }

    public void replaceChildComponent(Widget oldComponent, Widget newComponent) {
        if (!hasChildComponent(oldComponent)) {
            return;
        }
        getChildren().insert(newComponent, getChildren().indexOf(oldComponent));
        getChildren().remove(oldComponent);
        getElement().insertBefore(newComponent.getElement(),
                oldComponent.getElement());
        getElement().removeChild(oldComponent.getElement());
        orphan(oldComponent);
        adopt(newComponent);
        widgetToCell.remove(oldComponent);
        ChildCell cell = new ChildCell(newComponent, this);
        widgetToCell.put(newComponent, cell);
    }

    public void updateCaption(Paintable component, UIDL uidl) {
        // TODO Use this same class to display component captions
        // TODO I need to make this class usable without the server-side
    }

    /**
     * Only pixel values are accepted.
     */
    @Override
    public void setWidth(String w) {
        // Assume pixel values are always passed from ApplicationConnection

        int oldW = width;

        if (w != null && w != "") {
            int width = CSSUtil.parsePixel(w) - margin[1] - margin[3]
                    - border[1] - border[3] - padding[1] - padding[3];
            if (width < 0) {
                width = 0;
            }
            this.width = width;
            super.setWidth(width + "px");
        } else {
            width = -1;
            super.setWidth("");
            updateLayout(true);
        }
        updateActualSize();

        if (width != oldW) {
            if (isRendering) {
                sizeHasChangedDuringRendering = true;
            }
            updateLayout(false);
            client.runDescendentsLayout(this);
        }
    }

    /**
     * Only pixel values are accepted.
     */
    @Override
    public void setHeight(String h) {
        // Assume pixel values are always passed from ApplicationConnection

        int oldH = height;

        if (h != null && h != "") {
            int height = CSSUtil.parsePixel(h) - margin[0] - margin[2]
                    - border[0] - border[2] - padding[0] - padding[2];
            if (height < 0) {
                height = 0;
            }
            this.height = height;
            super.setHeight(height + "px");
        } else {
            height = -1;
            super.setHeight("");
            updateLayout(true);
        }

        if (height != oldH) {
            if (isRendering) {
                sizeHasChangedDuringRendering = true;
            }
            updateLayout(false);
            client.runDescendentsLayout(this);
        }
    }

    public void updateFromUIDL(UIDL uidl, ApplicationConnection client) {
        this.client = client;

        // Only non-cached UIDL messages can introduce changes
        if (uidl.getBooleanAttribute("cached")) {
            return;
        }

        useSpacing = uidl.hasAttribute("spacing");

        // Margins need to be set before any size calculations
        int bitMask = uidl.getIntAttribute("margins");
        if (marginInfo == null || marginInfo.getBitMask() != bitMask) {
            marginInfo = new VMarginInfo(bitMask);
            updateMargins();
        } else if (marginInfo.getBitMask() == bitMask) {
            // interpret as requestRepaint requested. Could be just an update on
            // margins or expand ratios, but we need to update the margin,
            // border and padding values in this case
            measureCSSAndSpacing();
            // Otherwise margins and spacing are measured in
            // #setStyleName(String), inside client.updateComponent
        }

        if (client.updateComponent(this, uidl, true)) {
            return;
        }

        horizontal = uidl.hasAttribute("horizontal");

        isRendering = true;

        updateDynamicSizeInfo(uidl);

        // If the layout has dynamic size, we need to update all relative sized
        // components after all components have been rendered. Alignments will
        // also need to be updated.
        final ArrayList<ChildCell> updateAfter = new ArrayList<ChildCell>();

        // Iterate through Paintables in UIDL, add new ones and remove any
        // old ones.
        final int uidlCount = uidl.getChildCount();
        final int layoutCount = getChildren().size();
        int uidlPos = 0;
        int layoutPos = 0;

        // Additional info that needs to be passed to components
        final ValueMap alignments = uidl.getMapAttribute("alignments");
        final ValueMap expandRatios = uidl.getMapAttribute("expandRatios");

        if (expandRatios.getKeySet().size() > 0) {
            compoundRatio = 0;
        } else {
            compoundRatio = -1;
        }

        for (; (uidlPos < uidlCount || layoutPos < layoutCount); uidlPos++) {

            final UIDL childUIDL = (uidlPos < uidlCount) ? uidl
                    .getChildUIDL(uidlPos) : null;

            final Widget uidlWidget = childUIDL != null ? (Widget) client
                    .getPaintable(childUIDL) : null;

            final Widget layoutWidget = (layoutPos < layoutCount) ? getChildren()
                    .get(layoutPos)
                    : null;

            // layout position is past UIDL position, no need to continue
            // (remaining old widgets are removed after this loop)
            if (uidlPos >= uidlCount && layoutPos < layoutCount) {
                break;
            }

            // Old widget in old location, all OK
            if (uidlWidget == layoutWidget) {
                layoutPos++;
            }

            // Widget is either new or has changed place

            else if (getChildren().contains(uidlWidget)) {
                // An old component has changed place
                if (getChildren().indexOf(uidlWidget) != layoutPos) {
                    // Detach from old position child.
                    uidlWidget.removeFromParent();

                    // Logical attach.
                    getChildren().insert(uidlWidget, layoutPos);

                    getElement().insertBefore(uidlWidget.getElement(),
                            getElement().getChildNodes().getItem(layoutPos));

                    adopt(uidlWidget);
                    layoutPos++;
                }
            } else {
                // A completely new widget is inserted to this position
                ChildCell cell = new ChildCell(uidlWidget, this);
                widgetToCell.put(uidlWidget, cell);

                // Logical attach
                getChildren().insert(uidlWidget, layoutPos);

                // Avoid inserts (they are slower than appends)
                if (layoutPos < getChildren().size() - 1) {
                    getElement().insertBefore(uidlWidget.getElement(),
                            getElement().getChildNodes().getItem(layoutPos));
                } else {
                    getElement().appendChild(uidlWidget.getElement());
                }
                layoutPos++;
                // Adopt
                adopt(uidlWidget);
            }
            ChildCell cell = widgetToCell.get(uidlWidget);
            cell.updateSizeInfo(childUIDL);
            cell.updateSpace();

            ((Paintable) uidlWidget).updateFromUIDL(childUIDL, client);

            if (alignments.containsKey(childUIDL.getId())) {
                cell.setAlignment(alignments.getInt(childUIDL.getId()));
            }
            if (expandRatios.containsKey(childUIDL.getId())) {
                final float ratio = expandRatios.getInt(childUIDL.getId());
                compoundRatio += ratio;
                cell.setExpandRatio(ratio);
            }

        } // All UIDL widgets painted

        // All remaining widgets are removed
        removeChildrenAfter(layoutPos);

        updateLayout(false);

        isRendering = false;

        if (sizeHasChangedDuringRendering) {
            updateLayout(true);
        }

    }

    protected void removeChildrenAfter(int pos) {
        int toRemove = getChildren().size() - pos;
        while (toRemove-- > 0) {
            Widget child = getChildren().get(pos);
            widgetToCell.remove(child);
            remove(child);
            client.unregisterPaintable((Paintable) child);
        }
    }

    protected void updateLayout(boolean reset) {
        // Hide overflows for the time of layouting
        getElement().getStyle().setProperty("overflow", "hidden");

        consumedSpace = 0;
        int totalSize = 0;
        int biggestSize = 0;
        final ArrayList<ChildCell> updateAfter = new ArrayList<ChildCell>();
        for (Widget w : getChildren()) {
            final ChildCell cell = getCells().get(w);
            if (reset) {
                // Reset widget size as well (true)
                cell.reset(true);
            }
            if (cell.updateAfterOtherCells()) {
                updateAfter.add(cell);
                if (!cell.isRelativeSizeInParentOrientation()) {
                    cell.updateWidgetMarginAndSize();
                    totalSize += isHorizontal() ? cell.getWidgetSize()
                            .getWidth() : cell.getWidgetSize().getHeight();
                }
            } else {
                cell.updateWidgetMarginAndSize();
                cell.updateSpace();
                cell.reAlign();
                totalSize += cell.getMaxSizeInParentOrientation();
                final int size = cell.getMaxSizeInNonParentOrientation();
                if (size > biggestSize) {
                    biggestSize = size;
                }
            }
        }

        if (useSpacing) {
            totalSize += spacing * (getChildren().size() - 1);
        }

        // Update consumed space before calculating expand ratio sizes
        consumedSpace = totalSize;

        for (int i = 0; i < updateAfter.size(); i++) {
            ChildCell cell = updateAfter.get(i);
            cell.updateWidgetMargin();
            cell.updateSpace();

            if (cell.hasRelativeSize()) {
                client.handleComponentRelativeSize(cell.getWidget());
                cell.updateWidgetSize();
            }

            cell.reAlign();

            final int size = cell.getMaxSizeInNonParentOrientation();
            if (size > biggestSize) {
                biggestSize = size;
            }

            if (cell.isRelativeSizeInParentOrientation()
                    || cell.getExpandRatio() != -1) {
                totalSize -= isHorizontal() ? cell.getWidgetSize().getWidth()
                        : cell.getWidgetSize().getHeight();
            }
            totalSize += cell.getMaxSizeInParentOrientation();
        }

        if (updateAfter.size() > 1) {
            ChildCell last = updateAfter.get(updateAfter.size() - 1);
            // Add rounding error pixel to last child (if more than one found,
            // otherwise no rounding errors should exhibit)
            totalSize -= last.getMaxSizeInParentOrientation();
            if (isHorizontal()) {
                last.getRenderSpace().setWidth(width - totalSize);
            } else {
                last.getRenderSpace().setHeight(height - totalSize);
            }
            if (last.hasRelativeSize()) {
                client.handleComponentRelativeSize(last.getWidget());
                last.updateWidgetSize();
            }
            last.reAlign();
            totalSize += last.getMaxSizeInParentOrientation();
        }

        if (undefWidth) {
            width = isHorizontal() ? totalSize : biggestSize;
            super.setWidth(width + "px");

        }
        if (undefHeight) {
            height = isHorizontal() ? biggestSize : totalSize;
            super.setHeight(height + "px");
        }

        getElement().getStyle().setProperty("overflow", "");
    }

    public boolean requestLayout(Set<Paintable> children) {
        return false;
        // FIXME
        // for (Paintable child : children) {
        // ChildCell cell = widgetToCell.get(child);
        // cell.reset(false);
        // cell.updateSpace();
        // cell.updateNaturalDimensions();
        // cell.reAlign();
        // }
        // if (undefWidth) {
        // final int oldWidth = width;
        // updateActualSize();
        // if (width != oldWidth) {
        // for (Widget child : getChildren()) {
        // ChildCell cell = widgetToCell.get(child);
        // cell.reset(false);
        // cell.updateSpace();
        // if (cell.isRelativeWidth() || cell.isRelativeHeight()) {
        // client.handleComponentRelativeSize(child);
        // }
        // cell.updateNaturalDimensions();
        // cell.reAlign();
        // }
        // return false;
        // }
        // }
        // return true;
    }

}
